/**************************************************************************** 
*  Copyright 2009 C&H Technologies, Inc.  All Rights Reserved.					    * 
*****************************************************************************/
/****************************************************************************
*
*	m221_ansi.c  
*
*	ANSI-C source module for the M221
*
* This file contains high level functions to control a M221 
* This file is written entirely in ANSI-C so that is may be compiled in 
* any environment with a supporting ANSI-C compiler It uses the APIS 
* library to perform instrument communications.  
*
*	This file is organized such that modifications are easy.  All functions 
*	that might need to be modified are located at the end of this file.  
*	Comments are provided that explain tasks of the function and how it might 
*	be modified.  All other function should not need to be modified.
*
*	C&H Part Number:
*
*	Revision History
*	-	Original Release Version 1.0
*****************************************************************************/
#include <ansi_c.h>
#include "m221_ansi.h"

/**************************************************************************** 
*  Non Exported Defines
*****************************************************************************/
/*  Driver Revision  */
#define		M221_DRIVER_REV			"1.0"

/**************************************************************************** 
*  Error Checking Macros
****************************************************************************/
#ifndef checkParamRange
#define  checkParamRange(value, min, max, param)		if ((value > max) || (value < min)) { \
																											error = M221_ERROR_PARAMETER1 + param-1;\
																											goto Error; \
																										} else 
#endif																				
																				

#ifndef checkParamPointer
#define  checkParamPointer(pointer, param)		if (pointer == NULL) { \
																							error = M221_ERROR_PARAMETER1 + param-1;\
																							goto Error; \
																							} else 
#endif																				

																				
#ifndef checkErr
#define checkErr(fCall)      if (error = (fCall), (error = (error < 0) ? error : M221_SUCCESS)) \
                                 {goto Error;}  else
#endif

#ifndef apisCheckErr
#define apisCheckErr(fCall)      if (error = (fCall), (error = (error != 0) ? error : M221_SUCCESS)) \
                                 {	handle->apis_error = error;\
																		error = M221_ERROR_APIS; \
																		goto Error;}  else
#endif

#ifndef apisCheckErrInit
#define apisCheckErrInit(fCall)      if (error = (fCall), (error = (error != 0) ? error : M221_SUCCESS)) \
                                 {	(*handle)->apis_error = error;\
																		error = M221_ERROR_APIS; \
																		goto Error;}  else
#endif




/**************************************************************************** 
*  Constants
*****************************************************************************/
static const char *errorMessageTable[M221_SPECIFIC_ERROR_END - M221_SPECIFIC_ERROR_BASE] = 
						{	"Timeout expired waiting for operation to complete",
							"An invalid parameter was passed to the function in parameter 1",
							"An invalid parameter was passed to the function in parameter 2",
							"An invalid parameter was passed to the function in parameter 3",
							"An invalid parameter was passed to the function in parameter 4",
							"The module returned an invalid ID",
							"Invalid format for the channel list was passed to the function",
							"An error occurred in the APIS Platform Support Module.  Use the function ma204_GetApisError() to retrieve the APIS error code",
							"The operation could not be completed because the system is out of memory"

						};

static const char *warningMessageTable[M221_SPECIFIC_WARNING_END - M221_SPECIFIC_WARNING_BASE]; 

/**************************************************************************** 
*  Non Exported Functions
*****************************************************************************/
static Int32 m221_VerifyList(Char *channel_list);
static Int32 m221_IDWriteByte(M221_HANDLE handle, UInt8 value);
static Int32 m221_IDWriteBit(M221_HANDLE handle, Bool value);
static Int32 m221_IDReadByte(M221_HANDLE handle, UInt8 *value);
static Int32 m221_IDReadBit(M221_HANDLE handle, Bool *value);


/**************************************************************************** 
* m221_VerifyList()
*
* IN: Char* channel_list
* 
*	Returns error upon finding an invalid list
*****************************************************************************/
static Int32 m221_VerifyList(Char *channel_list)
{
	Int32 error = M221_SUCCESS;
	Int32 start_channel, end_channel;
	Char *tokenptr;
	UInt8 num_items = 0;
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			num_items = sscanf(tokenptr, "%d", &start_channel);
			if( (start_channel < 0) || (start_channel > M221_LAST_CH_NUM) )
			{
				error = M221_ERROR_INVALID_CHANNEL_LIST;
				goto Error;
			}
		}
		else
		{
			num_items = sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
			if( (start_channel < 0) || (start_channel > M221_LAST_CH_NUM) || (end_channel < 0) || (end_channel > M221_LAST_CH_NUM) )
			{
				error = M221_ERROR_INVALID_CHANNEL_LIST;
				goto Error;
			}
		}
			
		if(num_items == 0)
		{
			error = M221_ERROR_INVALID_CHANNEL_LIST;
			goto Error;
		}
			
		tokenptr = strtok(NULL, ",");
			
	}
		
Error:
		return error;
}

/*****************************************************************************
*	ma221_IDReadBit
*
*	Reads a single bit from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m221_IDReadBit(M221_HANDLE handle, Bool *value)
{	
	Int32 error = M221_SUCCESS;
	UInt16 data;

/*  clock low  */ 	
	checkErr(m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, 0x4));
	
/*  clock high  */	
	checkErr(m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, 0x6));
	
/*  Read data  */	
	checkErr(m221_ReadRegister(handle, M221_ID_ACCESS_REGISTER, &data));
	*value = (Bool)data;

Error:	
	return error;
}


/*****************************************************************************
*	m221_IDWriteBit
*
*	Write a single bit to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m221_IDWriteBit(M221_HANDLE handle, Bool value)
{	
	Int32 error = M221_SUCCESS;

/*  data with clock low  */
	checkErr(m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, (0x4 | (value & 0x1))));
/*  data with clock high  */	 
	checkErr(m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, (0x6 | (value & 0x1))));

	Error:	
	return error;
}



/*****************************************************************************
*	m221_IDWriteByte
*
*	Write a command byte to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m221_IDWriteByte(M221_HANDLE handle, UInt8 value)
{	
	Int32 error = M221_SUCCESS;
	Int32 i;
	
/*  Initialize  */	
	checkErr(m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, 0));
	checkErr(m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, 0x4));
	
/*  Write start bit  */	
	checkErr(m221_IDWriteBit(handle, 1));
	for(i=0; i<=7; i++)
	{
	/*  Write data bits  */		
		checkErr(m221_IDWriteBit(handle, ((value &0x80)>>7)));
		value = value << 1;
	}
Error:	
	return error;
}

/*****************************************************************************
*	m221_IDReadByte
*
*	Reads a byte from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m221_IDReadByte(M221_HANDLE handle, UInt8 *value)
{	
	Int32 error = M221_SUCCESS;
	Int16 i;
	Bool data;
	
	*value = 0;
	for(i=7; i>=0; i--)
	{ 
	/*  read data 1 bit at a time  */	
		checkErr(m221_IDReadBit(handle, &data));
		*value |= ((data & 0x1) << i);
	}
	
Error:	
	return error;
}



/*****************************************************************************
*				Exported Functions
*****************************************************************************/

/*****************************************************************************
*	m221_RevisionQuery
*
*	Returns the version number of this driver.  
******************************************************************************/
Int32 m221_RevisionQuery(Int8 driver_rev[])
{	Int32 error = M221_SUCCESS;

/*  Check parameters  */
	checkParamPointer(driver_rev, 2);
	
	sprintf(driver_rev, "%s", M221_DRIVER_REV);
Error:
	return error;
}


/*****************************************************************************
* m221_ErrorMessage()
*
* Translates error codes to user readable error message
******************************************************************************/
Int32 m221_ErrorMessage(Int32 statusCode, Char errorMessage[])
{ 
	Int32 error = M221_SUCCESS;

/*  Check parameters  */
	checkParamPointer(errorMessage, 3);

/*  Check driver defined error codes  */
	if((statusCode >= M221_SPECIFIC_ERROR_BASE) && (statusCode < M221_SPECIFIC_ERROR_END))
	{	
		strcpy(errorMessage, errorMessageTable[statusCode - M221_SPECIFIC_ERROR_BASE]);
	}
	
	else if((statusCode >= M221_SPECIFIC_WARNING_BASE) && (statusCode < M221_SPECIFIC_WARNING_END))
	{	
		strcpy(errorMessage, warningMessageTable[statusCode - M221_SPECIFIC_WARNING_BASE]);
	}
	
	else if (statusCode == 0)
		sprintf(errorMessage, "No Error Occured");
	else
		sprintf(errorMessage, "Unknown Error Code [0x%x]", statusCode);
		
		
Error:
	return error;
}



/*****************************************************************************
*	m221_GetIdentWord
*
*	Reads the specified offset from the module's ID space and returns
*	the result.
******************************************************************************/
Int32 m221_GetIdentWord (M221_HANDLE handle, UInt16 offset, UInt16 *value)
{	Int32 error = M221_SUCCESS;
	UInt8 data;

/*  Check parameters  */
	checkParamPointer(value, 3);

	*value = 0;
	
/*  Send the read command to the serial EEPROM  */	
	checkErr(m221_IDWriteByte(handle, 0x80 | offset));
	
/*  Read upper byte  */	
	checkErr(m221_IDReadByte(handle, &data));
	*value = data << 8;
	
/*  Read lower byte  */	
	checkErr(m221_IDReadByte(handle, &data));
	*value |= data;
	
Error:	
	m221_WriteRegister(handle, M221_ID_ACCESS_REGISTER, 0);
	return error;
}


/******************************************************************************
* m221_RouteOpen
*																																				     
* Opens relay indicated by channel
*******************************************************************************/
Int32 m221_RouteOpen(M221_HANDLE handle, UInt16 channel)
{ 
	Int32 error = M221_SUCCESS;
	UInt16 bit_mask;
	UInt16 reg;
	Bool op_complete = 0;
	UInt16 t0 = 0, t_elapsed = 0, timeout = 0;
	INT8 time_char[8];
	
/*  Check parameters  */
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, M221_LAST_CH_NUM, 2);
	
/*	Initialize bit_mask and offset		*/
	bit_mask = 0x00;
	bit_mask |= (0x01 << channel);

	checkErr(m221_ReadRegister (handle, M221_RELAY_REGISTER, &reg));
	reg &= ~bit_mask;
	checkErr(m221_WriteRegister (handle, M221_RELAY_REGISTER, reg));

	if(handle->overlap == 0)
	{
		t0 = clock();
		timeout = handle->timeout;
			
		while( (!op_complete) && (t_elapsed <= timeout) ) 
		{ 
			m221_OpcQuery(handle, &op_complete);  // If overlap = 0, do not return until busy bit (0 -> 1) , or
			t_elapsed = clock() - t0;							//  	elapsed_time >= timeout.
		}
		if(t_elapsed >= timeout) 
		{	
			error = M221_ERROR_WAIT_TIMEOUT_EXPIRED;
			goto Error;
		}
	}

Error:
	return error;

}



/******************************************************************************
*	m221_RouteOpenQuery()			
*
*	Returns 1 if channel Opened (COM-NC) and 0 if channel Closed (COM-NO)
*******************************************************************************/
Int32 m221_RouteOpenQuery(M221_HANDLE handle, Int16 channel, Bool *opened)
{
	Int32 error = M221_SUCCESS;
	UInt16 bit_mask, reg;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, M221_LAST_CH_NUM, 2);
	checkParamPointer(opened, 3);
	
	bit_mask = 0;
	bit_mask |= (0x01 << channel);
	          									
	checkErr(m221_ReadRegister (handle, M221_RELAY_REGISTER, &reg));
	
	if(reg & bit_mask)
		*opened = M221_FALSE;
	else
		*opened = M221_TRUE;
	
Error:
	return error;
}


/******************************************************************************
* m221_RouteOpenListQuery
*	
*	Returns a comma seperated list of 1 (open) or 0 (close) for each channel
*	in the channel_list.
*******************************************************************************/
Int32 m221_RouteOpenListQuery(M221_HANDLE handle, Char *channel_list, Char *route_open_list)
{
	Int32 error = M221_SUCCESS;
	UInt16 reg, offset;
	Int32 chan, start_channel, end_channel;
	Bool opened;
	char *tokenptr;
	char str[20];
	
	checkParamPointer(handle, 1);
	checkParamPointer(channel_list, 2);
	checkParamPointer(route_open_list, 3);
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
				
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m221_RouteOpenQuery(handle, chan, &opened));
		
			// if (opened), append 1-comma to list, otherwise append 0-comma
			if(opened)
				strcat(route_open_list, "1,");
			else
				strcat(route_open_list, "0,");
		}
						
		tokenptr = strtok(NULL, ",");
	}

Error:
	return error;
}



/****************************************************************************
* m221_RouteCloseList()
*
* Closes a list of channels in comma and colon separated format
*  a colon denotes a range of channels to be closed
*****************************************************************************/
Int32 m221_RouteCloseList(M221_HANDLE handle, Char *channel_list)
{
	Int32 error = M221_SUCCESS;
	Int32 chan, start_channel, end_channel;
	Char *tokenptr;
	Char *channel_list2;
	
	channel_list2 = malloc(strlen(channel_list) + 1);

	strcpy(channel_list2, channel_list);
	checkErr(m221_VerifyList(channel_list2));  // Verify list first
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
				
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m221_RouteClose(handle, chan));
		}
					
		tokenptr = strtok(NULL, ",");
			
	}
		
Error:
		return error;
}



/**************************************************************************
* m221_RouteOpenList
*																																				
* Opens a list of channels passed as a comma separated string (char *list)
***************************************************************************/
Int32 m221_RouteOpenList(M221_HANDLE handle, Char *channel_list)
{
	Int32 error = M221_SUCCESS;
	Int32 chan, start_channel, end_channel;
	Char *tokenptr;
	Char *channel_list2;
	
	channel_list2 = malloc(strlen(channel_list) + 1);

	strcpy(channel_list2, channel_list);
	checkErr(m221_VerifyList(channel_list2));  // Verify list first
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
				
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m221_RouteOpen(handle, chan));
		}
					
		tokenptr = strtok(NULL, ",");
			
	}
		
Error:
		return error;
}


/***********************************************************************
* m221_RouteClose
*																																	             
* Closes (COM-NO) relay indicated by channel parameter
************************************************************************/
Int32 m221_RouteClose(M221_HANDLE handle, UInt16 channel)
{ 
	Int32 error = M221_SUCCESS;
	UInt16 bit_mask, reg;
	Bool op_complete = 0;
	UInt16 t0 = 0, t_elapsed = 0, timeout = 0;
	char time_char[8];
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, M221_LAST_CH_NUM, 2);
	
/*	Initialize bit_mask and offset	*/
	bit_mask = 0x00;
	bit_mask |= (0x01 << channel);

	checkErr(m221_ReadRegister (handle, M221_RELAY_REGISTER, &reg));
	reg |= bit_mask;
	checkErr(m221_WriteRegister (handle, M221_RELAY_REGISTER, reg));

	if(handle->overlap == 0)
	{
		t0 = clock();
		timeout = handle->timeout;
			
		while( (!op_complete) && (t_elapsed <= timeout) ) 
		{ 
			m221_OpcQuery(handle, &op_complete);  // If overlap = 0, do not return until busy bit (0 -> 1) , or
			t_elapsed = clock() - t0;							//  	elapsed_time >= timeout.
		}
		if(t_elapsed >= timeout) 
		{	
			error = M221_ERROR_WAIT_TIMEOUT_EXPIRED;
			goto Error;
		}
	}
	
Error:
	return error;

}


/***********************************************************************
* m221_RouteCloseQuery  
*
*	Returns 1 if Closed (COM-NO) and a 0 if Open (COM-NC)
************************************************************************/
Int32 m221_RouteCloseQuery(M221_HANDLE handle, Int16 channel, Bool *closed)
{
	Int32 error = M221_SUCCESS;
	UInt16 bit_mask, reg, result;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, M221_LAST_CH_NUM, 2);
	
	result = 0;
	bit_mask = 0;
	
	bit_mask |= (0x01 << channel);
	          									
	checkErr(m221_ReadRegister (handle, M221_RELAY_REGISTER, &reg));
	
	if(reg & bit_mask)
		*closed = M221_TRUE;
	else
		*closed = M221_FALSE;
	
Error:
	return error;
}


/***********************************************************************
* m221_RouteCloseListQuery
*
*	0 Opens the channel / 1 Closes the channel
************************************************************************/
Int32 m221_RouteCloseListQuery(M221_HANDLE handle, Char *channel_list, Char *route_close_list)
{
	Int32 error = M221_SUCCESS;
	UInt16 reg, offset;
	Int32 chan, start_channel, end_channel;
	Bool closed;
	char *tokenptr;
	char str[20];
	
	checkParamPointer(handle, 1);
	checkParamPointer(channel_list, 2);
	checkParamPointer(route_close_list, 3);
	
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
					
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m221_RouteCloseQuery(handle, chan, &closed));
	
			// if (closed), append 1-comma to list, otherwise append 0-comma
			if(closed)
				strcat(route_close_list, "1,");
			else
				strcat(route_close_list, "0,");
		}
						
		tokenptr = strtok(NULL, ",");
	}

Error:
	return error;
}

/***********************************************************************
*	m221_SetAll()				 
*
*	This function sets the state of all 8 relays simultanously 
*	
*	0 Opens the channel (COM-NC) / 1 Closes the channel (COM-NO)
************************************************************************/
Int32 m221_SetAll(M221_HANDLE handle, Bool ch7, Bool ch6, Bool ch5, Bool ch4, Bool ch3, Bool ch2, Bool ch1, Bool ch0)
{
	Int32 error = M221_SUCCESS;
	UInt16 bit_mask;
	Bool op_complete = 0;
	UInt16 t0 = 0, t_elapsed = 0, timeout = 0;
	char time_char[8];
	
	checkParamPointer(handle, 1);
	
	bit_mask = (ch7 << 7) | (ch6 << 6) | (ch5 << 5) | (ch4 << 4) | (ch3 << 3) | (ch2 << 2) | (ch1 << 1) | ch0;
	checkErr(m221_WriteRegister (handle, M221_RELAY_REGISTER, bit_mask));

	if(handle->overlap == 0)
	{
		t0 = clock();
		timeout = handle->timeout;
			
		while( (!op_complete) && (t_elapsed <= timeout) ) 
		{ 
			m221_OpcQuery(handle, &op_complete);  // If overlap = 0, do not return until busy bit (0 -> 1) , or
			t_elapsed = clock() - t0;							//  	elapsed_time >= timeout.
		}
		if(t_elapsed >= timeout) 
		{	
			error = M221_ERROR_WAIT_TIMEOUT_EXPIRED;
			goto Error;
		}
	}
	
Error:
	return error;
}



/***********************************************************************
*	m221_SetAllQuery     				 
*
* Returns 1 for relay closed (COM-NO) and 0 for relay opened (COKM-NC)
************************************************************************/
Int32 m221_SetAllQuery(M221_HANDLE handle, Bool *ch7, Bool *ch6, Bool *ch5, Bool *ch4, Bool *ch3, Bool *ch2, Bool *ch1, Bool *ch0)
{
	Int16 i, channel;
	Int32 error = M221_SUCCESS;
	Bool* ptr_array[8];
	Bool status = 0;
	
	checkParamPointer(handle, 1);
	
	ptr_array[0] = ch0; ptr_array[1] = ch1; ptr_array[2] = ch2; ptr_array[3] = ch3;
	ptr_array[4] = ch4; ptr_array[5] = ch5; ptr_array[6] = ch6; ptr_array[7] = ch7;
	
	for(i=0; i<=M221_LAST_CH_NUM; i++)
	{
		channel = i;
		checkErr(m221_RouteCloseQuery(handle, channel, &status));
		if(status)
			*ptr_array[i] = 1;
		else
			*ptr_array[i] = 0;
	}
	
Error:
	return error;
}



/***********************************************************************
*	m221_ChannelStatus
*
* Find status of channel bit in relay reg (0 or non-zero)
*
*	0 Channel Opened (COM-NC) / 1 Channel Closed (COM-NO)
************************************************************************/
Int32 m221_ChannelStatus(M221_HANDLE handle, UInt16 channel, Bool *status)
{ 
	Int32 error = M221_SUCCESS;
	UInt16 bit_mask, reg, channel_status;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, M221_LAST_CH_NUM, 2);
	checkParamPointer(status, 3);
	
	
	bit_mask = 0x00;
	bit_mask |= (0x01 << channel);
	          									
	checkErr(m221_ReadRegister (handle, M221_RELAY_REGISTER, &reg));
	
	if(reg & bit_mask)
		*status = M221_TRUE;
	else
		*status = M221_FALSE;
	
Error:
	return error;
}


/***********************************************************************
* m221_GetAttribute()
*
* Gets attribute value of	timeout and overlap
************************************************************************/
Int32 m221_GetAttribute(M221_HANDLE handle, Int16 attributeName, Int32 *attributeValue)
{ 
	Int32 error = M221_SUCCESS;
	checkParamPointer(handle, 1);
	checkParamPointer(attributeValue, 3);

	switch(attributeName)
	{
		case M221_OVERLAP_MODE:
			*attributeValue = handle->overlap;
			break;
			
		case M221_TIMEOUT:
			*attributeValue = handle->timeout;
			break;
	}
		
Error:
	return error;
}


/***********************************************************************
* m221_SetAttribute()
*
* Sets attributes of timeout & overlap
************************************************************************/
Int32 m221_SetAttribute( M221_HANDLE handle, Int16 attributeName, Int32 attributeValue)
{ 
	Int32 error = M221_SUCCESS;

	checkParamPointer(handle, 1);

	switch(attributeName)
	{
		case M221_OVERLAP_MODE:
			checkParamRange(attributeValue, 0, 1, 3);
			handle->overlap = attributeValue;
			break;
			
		case M221_TIMEOUT:
			handle->timeout = attributeValue;
			break;
			
	}

Error:
	return error;
}


/***********************************************************************
* m221_DiagTest()
*
* Tests for possible hardware problems
*   -returns test_result and message
************************************************************************/
Int32 m221_DiagTest(M221_HANDLE handle, Int16 *test_result, Char test_message[])
{ 
	Int32 error = M221_SUCCESS;

	UInt16 data, t0, t1, t_diff, time_out, j;


	checkParamPointer(handle, 1);
	checkParamPointer(test_message, 3);
	checkParamPointer(test_result, 2);

  *test_result = 0;
  strcpy(test_message, "DIAG_TEST_PASSED");
  
  /* reset disables interrupts*/
  
  /* verify busy bit */
	checkErr(m221_WriteRegister(handle, M221_RELAY_REGISTER, 0));
	t0 = clock();
	t1 = clock(); // Get time in milliseconds
	time_out = handle->timeout;
	t_diff = 0;
 	while ( t_diff < time_out )
  {
		t1 = clock();
		t_diff = t1 - t0;
    checkErr(m221_ReadRegister( (M221_HANDLE) handle, M221_STATUS_REGISTER, &data));
    if (!(data & M221_BUSY_STATUS_BIT))
     	goto passed1;
  }
			
stuck:
  *test_result = 1;
  sprintf(test_message, "MSG_BUSY_FAIL ON VERIFY BUSY BIT");
  goto done;
				
stuck2:
  *test_result = 1;
  sprintf(test_message, "MSG_BUSY_FAIL ON OPEN RELAYS");
  goto done;
				
stuck3:
  *test_result = 1;
  sprintf(test_message, "MSG_BUSY_FAIL ON CLOSE INDIVIDUAL RELAYS");
  goto done;
				
stuck4:
  *test_result = 1;
  sprintf(test_message, "MSG_BUSY_FAIL ON OPEN INDIVIDUAL RELAYS");
  goto done;
				
passed1:
    	
    	/* ---------  check data registers --------- */
            
    	/* first open all*/
 	if (m221_WriteWait((M221_HANDLE) handle, M221_RELAY_REGISTER, 0x00)) 
		goto stuck2;
 	
	checkErr(m221_ReadRegister( (M221_HANDLE) handle, M221_RELAY_REGISTER, &data));
  if ((data &0xff) != 0x00)
  {
 	  *test_result = 2;
 	  sprintf(test_message, "FAIL,RELAYS_REG" ,0,data);
 	  goto done;
  }
    	
  data = 0;
 	/* close 1 at a time */
 	for (j=0; j<=M221_LAST_CH_NUM; j++)
 	{
    /* send 1,2,4,8,16,32,64,128*/
    if (m221_WriteWait( (M221_HANDLE) handle, M221_RELAY_REGISTER, (UInt16)((1<<j)|data)) ) 
			goto stuck3;
  	
		checkErr(m221_ReadRegister( (M221_HANDLE) handle, M221_RELAY_REGISTER, &data));

		/* compare with 1,3,7,F,1F,3F,7F,FF */
    if ((data &0xFF) != (0xFF>>(M221_LAST_CH_NUM-j)) & 0xFF)
    {
	    *test_result = 2;
	    sprintf(test_message,"READBACK_FAILED AFTER CLOSING RELAYS");
	    goto done;
    }
	}
    
 	/* open 1 at a time */
 	for (j=0; j<=M221_LAST_CH_NUM; j++)
 	{
    /* send 1,2,4,8,16,32,64,128 */
    if (m221_WriteWait( (M221_HANDLE) handle, M221_RELAY_REGISTER, (UInt8)(0xFE<<j)) ) 
			goto stuck4;

		checkErr(m221_ReadRegister( (M221_HANDLE) handle, M221_RELAY_REGISTER, &data));

		/* compare with FE,FC,F8,F0,E,C,8,0 */
 	  if ( ((0xFE<<j)&0xFF) != (data & 0xFF))
    {
	    *test_result = 2;
	    sprintf(test_message,"READBACK_FAILED AFTER OPENING RELAYS");
		  goto done;
	  }
 	}
    	   
done:
  /* reset the card to beginning state */
  checkErr(m221_Reset((M221_HANDLE) handle));
Error:
		return error;
}


/**********************************************************************
* m221_WriteWait()
*
* used in diag test. writes and waits for busy bits without interrupts
***********************************************************************/
Int16 m221_WriteWait(M221_HANDLE handle, UInt16 reg, UInt16 data)
{ 

	Int32 error = M221_SUCCESS;
	UInt16 t0; // Time 
	
	checkParamPointer(handle, 1);


	t0 = clock();

	m221_WriteRegister(handle, M221_RELAY_REGISTER, data);
	for(;;)
	{
		m221_ReadRegister( (M221_HANDLE) handle, M221_STATUS_REGISTER, &data);
		if (!(data & M221_BUSY_STATUS_BIT))
			return 0;
	
		if ( (clock() - t0 ) > (handle->timeout) )
			return 1;
	}
Error:
	return error;
}


/***********************************************************************
* m221_OpcQuery()
*
* Returns 1 when operation complete
*
*  Note: Use when overlap = 0 in between
*        open / close relay operations to
*        insure settling time of relays has
*        elapsed. (~13 milliseconds). The
*        busy bit goes to 1 at this time
***********************************************************************/
Int32 m221_OpcQuery(M221_HANDLE handle, Int16 *state )
{ 
	Int32 error = M221_SUCCESS;
	UInt16 busy;
	
	checkParamPointer(handle, 1);
	checkParamPointer(state, 2);
	
	checkErr(m221_ReadRegister(handle, M221_STATUS_REGISTER, &busy));
		
	if (busy & M221_BUSY_STATUS_BIT)
		*state = 1;
	else
		*state = 0;

Error:
	return error;
}


/*****************************************************************************
*	m221_EnableInterrupts 
*
*	Configures the m221 to produce interrupts.  This function does not
*	cconfigure the carrier to pass interrupts on to the host controller
******************************************************************************/
Int32 m221_EnableInterrupts(M221_HANDLE handle, Bool enable)
{	
	Int32 error = M221_SUCCESS;
	UInt16 control_reg; 
	UInt16 data;
	
	checkParamPointer(handle, 1);
	checkParamRange(enable, 0, 1, 2);

	checkErr(m221_ReadRegister (handle, M221_CONTROL_REGISTER, &control_reg)); 
	control_reg &= ~M221_ENABLE_RELAY_INTERRUPT_BIT;						  
	data = (enable << 1) & M221_ENABLE_RELAY_INTERRUPT_BIT;        
	control_reg |= data;
	checkErr(m221_WriteRegister (handle, M221_CONTROL_REGISTER, control_reg));
	
Error:
	return error;
}

/*****************************************************************************
*	m221_InterruptStatus
*	
******************************************************************************/
Int32 m221_InterruptStatus(M221_HANDLE handle, Bool *status)
{	
	Int32 error = M221_SUCCESS;
	UInt16 stat_reg_value;
	
	checkParamPointer(handle, 1);
	checkParamPointer(handle, 2);
	
	checkErr(m221_ReadRegister (handle, M221_STATUS_REGISTER, &stat_reg_value));
	if(stat_reg_value & M221_INTERRUPT_STATUS_BIT)
		*status = M221_TRUE;
	else
		*status = M221_FALSE;

Error:
	return error;
}


/*****************************************************************************
*	m221_ClearInterrupts
*
*	Clears the interrupt pending
*	
******************************************************************************/
Int32 m221_ClearInterrupts (M221_HANDLE handle)
{	Int32 error = M221_SUCCESS;
	UInt16 stat_reg_value;

	checkErr(m221_ReadRegister(handle, M221_INTERRUPT_REGISTER, &stat_reg_value));
	// a read of this register clears the interrupt pending bit, ignore value
	
Error:
	return error;
}










/*****************************************************************************
*	Modify the functions below this point to use the M221 driver in an 
*	environment not supported by an APIS platform support module. All functions
*	above this point should not have to be modified.
*
*	In general, any call to an APIS function (apis_func()) will need to be
*	modified to communicate with the instrument or operating system being used.
*
******************************************************************************/
/*****************************************************************************
*	m221_Init
*
*	Opens a M221 session to the instrument specified by the resource 
*	description parameter.  If IDQuery is true then this function 
*	verifies that the instrument is a m221.  If reset is true then
*	this function resets the instrument to it's default state.
******************************************************************************/
Int32 m221_Init (Int32 instr, Bool IDQuery, Bool reset, M221_HANDLE *handle)
{	Int32 error = M221_SUCCESS;
	Int32 warning = M221_SUCCESS;
	UInt16 value;

/*  Check parameters  */
	checkParamRange(IDQuery, 0, 1, 2);
	checkParamRange(reset, 0, 1, 3);
	checkParamPointer(handle, 4);

/*  Allocate the data structure  */
  (*handle) = 0;
  (*handle) = malloc(sizeof(M221_INSTR));
	if((*handle) == NULL)
	{	error = M221_ERROR_OUT_OF_MEMORY;
		goto Error;
	}
		
	(*handle)->apis_handle = 0;
	apisCheckErrInit(apis_open((APIS_PATH)instr, &(*handle)->apis_handle));
	
/*  Initialize attributes  */
  (*handle)->overlap = 0;  
  (*handle)->timeout = 20;   // milliseconds   

/* Perform ID Query */
	if(IDQuery == M221_TRUE)
	{	
		checkErr(m221_GetIdentWord (*handle, 17, &value));
		value &= 0x0FFF;
		if (value != (UInt16)M221_VXI_MANF_ID) 
		{	error = M221_ERROR_FAIL_ID_QUERY;
			goto Error;
	  }

		checkErr(m221_GetIdentWord (*handle, 18, &value));
		value &= 0x0FFF;
		if (value != (UInt16)M221_VXI_MODEL_CODE) 
		{	error = M221_ERROR_FAIL_ID_QUERY;
			goto Error;
		} 
	}

/* Reset Instrument */
  if (reset == M221_TRUE) 
  {	
		checkErr(m221_Reset (*handle));
  }

Error:
/*  Do not free handle here.  The Calling function must call m221_Close upon receiving an error  */	

		if((error == M221_SUCCESS) && (warning != M221_SUCCESS))	// if no errors then return the warnings				
			error = warning;
							
		return error; 
}

/*****************************************************************************
*	m221_Close
*
*	Close the session to the instrument.
******************************************************************************/
Int32 m221_Close (M221_HANDLE handle)
{	Int32 error = M221_SUCCESS;

	if(handle->apis_handle != NULL)
		apisCheckErr(apis_close(handle->apis_handle));
	free(handle);
	
Error:
	return error;
}


/*****************************************************************************
*	m221_Reset 																												
*
******************************************************************************/
Int32 m221_Reset(M221_HANDLE handle)
{	
	Int32 error = M221_SUCCESS;
	UInt16 control_reg; 
	UInt16 data;
	
	checkParamPointer(handle, 1);
	
	checkErr(m221_ReadRegister (handle, M221_CONTROL_REGISTER, &control_reg));
	control_reg |= M221_RESET_BIT;
	checkErr(m221_WriteRegister (handle, M221_CONTROL_REGISTER, control_reg));
	control_reg &= ~M221_RESET_BIT;                                	
	checkErr(m221_WriteRegister (handle, M221_CONTROL_REGISTER, control_reg));
	
Error:
	return error;
}





/*****************************************************************************
*	m221_ReadRegister
*
*	Read the specified offset from the module's I/O space and returns
*	the result.
******************************************************************************/
Int32 m221_ReadRegister (M221_HANDLE handle, UInt16 offset, UInt16 *value)
{	Int32 error = M221_SUCCESS;

/*  Check parameters  */
	checkParamPointer(value, 3);
	
	apisCheckErr(apis_read(handle->apis_handle, 2, offset, value));
	
Error:
	return error;
}


/*****************************************************************************
*	m221_WriteRegister
*
*	Writes to the specified offset in the module's I/O space
******************************************************************************/
Int32 m221_WriteRegister (M221_HANDLE handle, UInt16 offset, UInt16 value)
{	
	Int32 error = M221_SUCCESS;
	
	apisCheckErr(apis_write(handle->apis_handle, 2, offset, value));

Error:
	return error;
}


/*****************************************************************************
*	m221_InstallIrq
*	
*	Installs the interrupts handler to be called when an interrupt occurs
*	and configures VISA to handle the interrupts by calling the handler.
******************************************************************************/
Int32 m221_InstallIrq(M221_HANDLE handle, void *interruptHandler, Int32 vector, Int32 level, Int32 mode, void *var_ptr)
{	Int32 error = M221_SUCCESS;

/*  Check parameters  */
	checkParamPointer(interruptHandler, 2);

	apisCheckErr(apis_irqinstall(handle->apis_handle, interruptHandler, vector, level, mode, var_ptr));

Error:
	return error;
}


/*****************************************************************************
*	m221_UninstallIrq
*
*	Un-installs the interrupt handler for the instrument and disables
*	VISA interrupt handeling for the instrument
******************************************************************************/
Int32 m221_UninstallIrq(M221_HANDLE handle)
{	Int32 error = M221_SUCCESS;

	apisCheckErr(apis_irqremove(handle->apis_handle));

Error:
	return error;
}


/*****************************************************************************
*	m221_GetApisError
*
*	Returns the saved apis error code
******************************************************************************/
Int32 m221_GetApisError (M221_HANDLE handle, UInt32 *apis_error)
{	Int32 error = M221_SUCCESS;
	*apis_error = handle->apis_error;
	return error;
}
